/* partition.h  -  Partition table handling */

/* Copyright 1992-1995 Werner Almesberger. See file COPYING for details. */


#ifndef PARTITION_H
#define PARTITION_H

void part_verify(int dev_nr);

/* Verify the partition table of the disk of which dev_nr is a partition. May
   also try to "fix" a partition table. */

#endif
